import astropy.units as u
import numpy as np
import xarray as xr

from psipy.model import base


def test_mas_model(mas_model):
    # Check that loading a single file works
    assert isinstance(mas_model, base.ModelOutput)
    assert "MAS output in directory" in str(mas_model)
    assert 'rho' in str(mas_model)

    rho = mas_model['rho']
    assert isinstance(rho, base.Variable)
    assert isinstance(rho.data, xr.DataArray)
    assert rho.unit == u.cm**-3
    assert rho.n_timesteps == 1
    assert str(rho) == """
Variable
--------
Name: rho
Grid size: (128, 111, 141) (phi, theta, r)
Timesteps: 1
"""


def test_persistance(mas_model):
    # Check that a variable requested twice only makes one copy of the data in
    # memory
    rho1 = mas_model['rho']
    rho2 = mas_model['rho']
    # This checks that rho1 and rho2 reference the same underlying data
    assert rho1 is rho2


def test_change_units(mas_model):
    # Check that loading a single file works
    rho = mas_model['rho']
    assert rho.unit == u.cm**-3
    old_data = rho._data.copy()
    rho.unit = u.m**-3
    assert rho.unit == u.m**-3
    assert np.allclose(rho._data.values, 1e6 * old_data.values)



"""
Field lines with pyvista
========================
Visualising traced field lines with pyvista.

pyvista is a Python package that provides a wrapper to the popular VTK library
for 3D visualisation. Unlike Matplotlib this is "true" 3D rendering, and is
much more performant in comparison.
"""
###############################################################################
# First, load the required modules.
import astropy.constants as const
import astropy.units as u
import numpy as np

from psipy.data import sample_data
from psipy.model import MASOutput
from psipy.tracing import FortranTracer
from psipy.visualization.pyvista import MASPlotter

###############################################################################
# Load a set of MAS output files.
mas_path = sample_data.mas_sample_data()
model = MASOutput(mas_path)

###############################################################################
# To trace field lines, start by creating a tracer. Then we create a set of
# seed points at which the field lines are drawn from.
tracer = FortranTracer()

nseeds = 20
r = np.ones(nseeds**2) * 40
lat = np.linspace(-45, 45, nseeds**2, endpoint=False) * u.deg
lon = np.random.rand(nseeds**2) * 360 * u.deg

# Do the tracing!
flines = tracer.trace(model, r=r, lat=lat, lon=lon)

###############################################################################
# Visualise the results

plotter = MASPlotter(model)
br = model['br']
for fline in flines:
    # Set color with polarity near the inner boundary
    color = br.sample_at_coords(np.mod(fline.lon[1], 2 * np.pi * u.rad),
                                fline.lat[1],
                                fline.r[1] * const.R_sun) > 0
    color = {0: 'red', 1: 'blue'}[int(color)]
    plotter.add_fline(fline, color=color)

# Add a sphere at the inner boundary of the model
plotter.add_sphere(np.min(br.r_coords))
plotter.show()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.projections.polar import ThetaFormatter


def clear_axes_labels(ax):
    """
    Remove labels from both x and y axes.
    """
    ax.set_xlabel('')
    ax.set_ylabel('')


def set_theta_formatters(ax):
    """
    Set both x and y axes to have theta formatters (ie. degrees)
    """
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_major_formatter(ThetaFormatter())
        axis.set_minor_formatter(ThetaFormatter())


def setup_radial_ax(ax):
    if ax is None:
        ax = plt.gca()
    return ax


def format_radial_ax(ax):
    ax.set_aspect('equal')
    ax.set_xlim(0, 2 * np.pi)
    ax.set_ylim(-np.pi / 2, np.pi / 2)
    clear_axes_labels(ax)

    # Tick label formatting
    set_theta_formatters(ax)
    ax.set_xticks(np.deg2rad(np.linspace(0, 360, 7, endpoint=True)))
    ax.set_yticks(np.deg2rad(np.linspace(-90, 90, 7, endpoint=True)))


def setup_polar_ax(ax):
    if ax is None:
        ax = plt.subplot(projection='polar')
    elif ax.name != 'polar':
        raise ValueError('ax must have a polar projection')
    return ax


def format_polar_ax(ax):
    # Plot formatting
    ax.set_rlim(0)
    ax.set_thetalim(-np.pi / 2, np.pi / 2)
    clear_axes_labels(ax)

    # Tick label formatting
    # Set theta ticks
    ax.set_xticks([])


def format_equatorial_ax(ax):
    # Plot formatting
    ax.set_rlim(0)
    ax.set_thetalim(0, 2 * np.pi)
    clear_axes_labels(ax)

    # Tick label formatting
    # Remove theta ticks
    ax.set_xticks([])


def animate_time(ax, slice, quad_mesh):
    """
    Animate *slice* over the *time* dimension.
    """
    n_timesteps = len(slice.coords['time'])

    def animate(frame_number):
        time_slice = slice.isel(time=frame_number)
        quad_mesh.set_array(time_slice.data.T)
        return quad_mesh

    return FuncAnimation(ax.figure, animate, frames=n_timesteps)



import astropy.constants as const
import astropy.units as u
import pytest

from psipy.model import Variable


def test_var_error(mas_model):
    with pytest.raises(RuntimeError, match='not in list of known variables'):
        mas_model['not_a_var']


def test_radial_normalised(mas_model):
    norm = mas_model['rho'].radial_normalized(-2)
    assert isinstance(norm, Variable)


# Check different shaped input, including lon/lat points that go up/down in
# value
@pytest.mark.parametrize(
    'lon, lat, r',
    [(1*u.deg, 1*u.deg, 30*const.R_sun),
     ([1, 2] * u.deg, [1, 2] * u.deg, [30, 31] * const.R_sun),
     ([1, 0] * u.deg, [1, 0] * u.deg, [30, 31] * const.R_sun),
     ])
def test_sample_at_coords_mas(mas_model, lon, lat, r):
    # Check scalar coords
    rho = mas_model['rho'].sample_at_coords(lon=lon, lat=lat, r=r)
    assert rho.unit == mas_model['rho'].unit
    assert u.allclose(rho[0], [447.02795493] * u.cm**-3)


@pytest.mark.parametrize(
    'lon, lat, r',
    [(1*u.deg, 1*u.deg, 1*const.R_sun),
     ([1, 2] * u.deg, [1, 2] * u.deg, [1, 1.01] * const.R_sun),
     ([1, 0] * u.deg, [1, 0] * u.deg, [1, 1.01] * const.R_sun),
     ])
def test_sample_at_coords_pluto(pluto_model, lon, lat, r):
    # Check scalar coords
    rho = pluto_model['rho'].sample_at_coords(lon=lon, lat=lat, r=r)
    assert rho.unit == pluto_model['rho'].unit
    assert u.allclose(rho[0], [13.50442343])



"""
Field lines with Matplotlib
===========================
Visualising traced field lines with Matplotlib.
"""
###############################################################################
# First, load the required modules.
import astropy.constants as const
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from psipy.data import sample_data
from psipy.model import MASOutput
from psipy.tracing import FortranTracer

###############################################################################
# Load a set of MAS output files.
mas_path = sample_data.mas_sample_data()
model = MASOutput(mas_path)

###############################################################################
# To trace field lines, start by creating a tracer. Then we create a set of
# seed points at which the field lines are drawn from.
tracer = FortranTracer()

nseeds = 5
# Radius
r = np.ones(nseeds**2) * 40
# Some points near the equatorial plane
lat = np.linspace(-10, 10, nseeds**2, endpoint=False) * u.deg
# Choose random longitudes
lon = np.random.rand(nseeds**2) * 360 * u.deg

# Do the tracing!
flines = tracer.trace(model, r=r, lat=lat, lon=lon)

###############################################################################
# flines is a list, with each item containing a field line object
print(flines[0])

###############################################################################
# To easily visualise the result, here we use Matplotlib. Note that Matplotlib
# is not a 3D renderer, so has several drawbacks (including performance).
fig = plt.figure()
ax = fig.add_subplot(projection='3d')

br = model['br']
for fline in flines:
    # Set color with polarity on the inner boundary
    color = br.sample_at_coords(np.mod(fline.lon[0], 2 * np.pi * u.rad),
                                fline.lat[0],
                                fline.r[0] * const.R_sun) > 0
    color = {0: 'red', 1: 'blue'}[int(color)]
    ax.plot(*fline.xyz, color=color, linewidth=2)

lim = 60
ax.set_xlim(-lim, lim)
ax.set_ylim(-lim, lim)
ax.set_zlim(-lim, lim)
plt.show()




import astropy.units as u
import numpy as np
import scipy.interpolate
import xarray as xr

from psipy.io import get_mas_variables, read_mas_file
from .base import ModelOutput

__all__ = ['MASOutput']


# A mapping from unit names to their units, and factors the data needs to be
# multiplied to get them into these units.
_vunit = [u.km / u.s, 481.37]
_bunit = [u.G, 2.205]
_junit = [u.A / u.m**2, 2.267e4]
_mas_units = {'vr': _vunit,
              'vt': _vunit,
              'vp': _vunit,
              'va': _vunit,
              'br': _bunit,
              'bt': _bunit,
              'bp': _bunit,
              'bmag': _bunit,
              'rho': [u.cm**-3, 1.67e-16 / 1.67e-24],
              't': [u.K, 2.804e7],
              'p': [u.Pa, 3.875717e-2],
              'jr': _junit,
              'jt': _junit,
              'jp': _junit
              }
_2pi = 2 * np.pi


class MASOutput(ModelOutput):
    """
    The results from a single run of MAS.

    This is a storage object that contains a number of `Variable` objects. It
    is designed to be used like::

        mas_output = MASOutput('directory')
        br = mas_output['br']

    Notes
    -----
    Variables are loaded on demand. To see the list of available variables
    use `MASOutput.variables`, and to see the list of already loaded variables
    use `MASOutput.loaded_variables`.
    """
    def get_unit(self, var):
        return _mas_units[var]

    def get_variables(self):
        return get_mas_variables(self.path)

    def load_file(self, var):
        return read_mas_file(self.path, var)

    def __repr__(self):
        return f'psipy.model.mas.MASOutput("{self.path}")'

    def __str__(self):
        return f"MAS output in directory {self.path}\n" + super().__str__()

    def cell_corner_b(self, t_idx=None):
        """
        Get the magnetic field vector at the cell corners.

        Parameters
        ----------
        t_idx : int, optional
            If more than one timestep is present in the loaded model, a
            timestep index at which to get the vectors must be provided.

        Returns
        -------
        xarray.DataArray

        Notes
        -----
        The phi limits go from 0 to 2pi inclusive, with the vectors at phi=0
        equal to the vectors at phi=2pi.
        """
        if not set(['br', 'bt', 'bp']) <= set(self.variables):
            raise RuntimeError('MAS output must have the br, bt, bp variables loaded')

        # Interpolate radial coordinate
        new_rcoord = self['bt'].r_coords
        br = scipy.interpolate.interp1d(self['br'].r_coords,
                                        self['br'].data.isel(time=t_idx or 0),
                                        axis=2,
                                        fill_value='extrapolate')(new_rcoord)

        # Interpolate theta coordinate
        new_tcoord = self['bp'].theta_coords
        bt = scipy.interpolate.interp1d(self['bt'].theta_coords,
                                        self['bt'].data.isel(time=t_idx or 0),
                                        axis=1,
                                        fill_value='extrapolate')(new_tcoord)

        # Interoplate phi coordinate
        new_pcoord = self['br'].phi_coords
        bp = scipy.interpolate.interp1d(self['bp'].phi_coords,
                                        self['bp'].data.isel(time=t_idx or 0),
                                        axis=0,
                                        fill_value='extrapolate')(new_pcoord)
        # Calculate edge/cyclic phi value
        old_pcoord = self['bp'].phi_coords
        edge_pcoord = [old_pcoord[-1],
                       old_pcoord[0] + _2pi]
        edge_data = self['bp'].data.isel(time=t_idx or 0)
        edge_data = np.stack([edge_data[-1, :, :],
                              edge_data[0, :, :]], axis=0)
        bp_edge = scipy.interpolate.interp1d(
            edge_pcoord, edge_data, axis=0)(_2pi)
        bp_edge = bp_edge.reshape((1, *bp_edge.shape))

        # Add an extra layer of cells at phi=2pi for the tracer
        br = np.concatenate((br, br[0:1]), axis=0)
        bt = np.concatenate((bt, bt[0:1]), axis=0)
        bp = np.concatenate((bp_edge, bp[1:, :, :], bp_edge), axis=0)
        new_pcoord = np.append(new_pcoord, _2pi)

        return xr.DataArray(np.stack([bp, bt, br], axis=-1),
                            dims=['phi', 'theta', 'r', 'component'],
                            coords=[new_pcoord, new_tcoord, new_rcoord,
                                    ['bp', 'bt', 'br']])

    def cell_centered_v(self, extra_phi_coord=False):
        """
        Get the velocity vector at the cell centres.

        Because the locations of the vector component outputs

        Parameters
        ----------
        extra_phi_coord: bool
            If `True`, add an extra phi slice.
        """
        if not set(['vr', 'vt', 'vp']) <= set(self.variables):
            raise RuntimeError('MAS output must have the vr, vt, vp variables loaded')

        # Interpolate new radial coordiantes
        new_rcoord = self['vr'].r_coords
        vt = scipy.interpolate.interp1d(self['vt'].r_coords,
                                        self['vt'].data,
                                        axis=2)(new_rcoord)
        vp = scipy.interpolate.interp1d(self['vp'].r_coords,
                                        self['vp'].data,
                                        axis=2)(new_rcoord)

        # Interpolate new theta coordinates
        new_tcoord = self['vt'].theta_coords
        vr = scipy.interpolate.interp1d(self['vr'].theta_coords,
                                        self['vr'].data,
                                        axis=1)(new_tcoord)
        vp = scipy.interpolate.interp1d(self['vp'].theta_coords,
                                        vp,
                                        axis=1)(new_tcoord)
        # Don't need to interpolate phi coords, but get a copy
        new_pcoord = self['vr'].phi_coords

        if extra_phi_coord:
            dphi = np.mean(np.diff(new_pcoord))
            assert np.allclose(new_pcoord[0] + 2 * np.pi, new_pcoord[-1] + dphi)

            new_pcoord = np.append(new_pcoord, new_pcoord[-1] + dphi)
            vp = np.append(vp, vp[0:1, :, :], axis=0)
            vt = np.append(vt, vt[0:1, :, :], axis=0)
            vr = np.append(vr, vr[0:1, :, :], axis=0)

        return xr.DataArray(np.stack([vp, vt, vr], axis=-1),
                            dims=['phi', 'theta', 'r', 'component'],
                            coords=[new_pcoord, new_tcoord, new_rcoord,
                                    ['vp', 'vt', 'vr']])



import copy
import textwrap

import astropy.constants as const
import astropy.units as u
import numpy as np
import xarray as xr
from scipy import interpolate

import psipy.visualization as viz
from psipy.util.decorators import add_common_docstring

__all__ = ['Variable']


# Some docstrings that are used more than once
quad_mesh_link = ':class:`~matplotlib.collections.QuadMesh`'
# TODO: fix this to ':class:`~matplotlib.animation.FuncAnimation`'
animation_link = 'animation'

returns_doc = textwrap.indent(f"""
{quad_mesh_link} or {animation_link}
    If a timestep is specified, the {quad_mesh_link} of the plot is returned.
    Otherwise an {animation_link} is returned.
""", '        ')


class Variable:
    """
    A single scalar variable.

    This class primarily contains methods for plotting data. It can be created
    with any `xarray.DataArray` that has ``['theta', 'phi', 'r', 'time']``
    fields.

    Parameters
    ----------
    data : xarray.Dataset
        Variable data.
    name : str
        Variable name.
    unit : astropy.units.Quantity
        Variable unit.
    """
    def __init__(self, data, name, unit):
        # Convert from xarray Dataset to DataArray
        self._data = data[name]
        # Sort the data once now for any interpolation later
        self._data = self._data.transpose(*['phi', 'theta', 'r', 'time'])
        self._data = self._data.sortby(['phi', 'theta', 'r', 'time'])
        self.name = name
        self._unit = unit

    def __str__(self):
        return textwrap.dedent(f'''
        Variable
        --------
        Name: {self.name}
        Grid size: {len(self.phi_coords), len(self.theta_coords), len(self.r_coords)} (phi, theta, r)
        Timesteps: {len(self.time_coords)}
        ''')

    @property
    def data(self):
        """
        `xarray.DataArray` with the data.
        """
        return self._data

    @property
    def unit(self):
        """
        Units of the scalar data.
        """
        return self._unit

    @unit.setter
    def unit(self, new_unit):
        # This line will error if untis aren't compatible
        conversion = float(1 * self._unit / new_unit)
        self._data *= conversion
        self._unit = new_unit

    @property
    def r_coords(self):
        """
        Radial coordinate values.
        """
        return self._data.coords['r'].values

    @r_coords.setter
    def r_coords(self, coords):
        self._data.coords['r'] = coords

    @property
    def theta_coords(self):
        """
        Latitude coordinate values.
        """
        return self._data.coords['theta'].values

    @property
    def phi_coords(self):
        """
        Longitude coordinate values.
        """
        return self._data.coords['phi'].values

    @property
    def time_coords(self):
        """
        Timestep coordinate values.
        """
        return self._data.coords['time'].values

    @property
    def n_timesteps(self):
        """
        Number of timesteps.
        """
        return len(self.time_coords)

    def radial_normalized(self, radial_exponent):
        r"""
        Return a radially normalised copy of this variable.

        Multiplies the variable by :math:`(r / r_{\odot})^{\gamma}`,
        where :math:`\gamma` = ``radial_exponent`` is the given exponent.

        Parameters
        ----------
        radial_exponent : float

        Returns
        -------
        Variable
        """
        r = self.data.coords['r']
        rsun_au = float(u.AU / const.R_sun)
        data = self.data * (r * rsun_au)**radial_exponent
        units = self.unit * (u.AU)**radial_exponent
        name = self.name + f' $r^{radial_exponent}$'
        return Variable(xr.Dataset({name: data}), name, units)

    # Methods for radial cuts
    @add_common_docstring(returns_doc=returns_doc)
    def plot_radial_cut(self, r_idx, t_idx=None, ax=None, **kwargs):
        """
        Plot a radial cut.

        Parameters
        ----------
        r_idx : int
            Radial index at which to slice the data.
        t_idx : int, optional
            Time index at which to slice the data. If not given, an anmiation
            will be created across all time indices.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to
            `xarray.plot.pcolormesh`.

        Returns
        -------
        {returns_doc}
        """
        r_slice = self.data.isel(r=r_idx)
        time_slice = r_slice.isel(time=t_idx or 0)

        # Setup axes
        ax = viz.setup_radial_ax(ax)
        # Set colorbar string
        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        quad_mesh = time_slice.plot(x='phi', y='theta', ax=ax, **kwargs)
        # Plot formatting
        r = r_slice['r'].values
        ax.set_title(f'{self.name}, r={r:.2f}' + r'$R_{\odot}$')
        viz.format_radial_ax(ax)

        if t_idx is not None or self.n_timesteps == 1:
            return quad_mesh
        else:
            return viz.animate_time(ax, r_slice, quad_mesh)

    def contour_radial_cut(self, r_idx, levels, t_idx=0, ax=None, **kwargs):
        """
        Plot contours on a radial cut.

        Parameters
        ----------
        r_idx : int
            Radial index at which to slice the data.
        levels : list
            List of levels to contour.
        t_idx : int, optional
            Time index at which to slice the data.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = viz.setup_radial_ax(ax)
        sliced = self.data.isel(r=r_idx, time=t_idx)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xr.plot.contour(sliced, x='phi', y='theta', ax=ax,
                        levels=levels, **kwargs)
        ax.set_title(title)
        viz.format_radial_ax(ax)

    @add_common_docstring(returns_doc=returns_doc)
    def plot_phi_cut(self, phi_idx, t_idx=None, ax=None, **kwargs):
        """
        Plot a phi cut.

        Parameters
        ----------
        phi_idx : int
            Index at which to slice the data.
        t_idx : int, optional
            Time index at which to slice the data. If not given, an anmiation
            will be created across all time indices.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to
            `xarray.plot.pcolormesh`.

        Returns
        -------
        {returns_doc}
        """
        phi_slice = self.data.isel(phi=phi_idx)
        time_slice = phi_slice.isel(time=t_idx or 0)

        ax = viz.setup_polar_ax(ax)
        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        # Take slice of data and plot
        quad_mesh = time_slice.plot(x='theta', y='r', ax=ax, **kwargs)
        viz.format_polar_ax(ax)

        phi = np.rad2deg(time_slice['phi'].values)
        ax.set_title(f'{self.name}, ' + r'$\phi$= ' + f'{phi:.2f}' +
                     r'$^{\circ}$')

        if t_idx is not None or self.n_timesteps == 1:
            return quad_mesh
        else:
            return viz.animate_time(ax, phi_slice, quad_mesh)

    def contour_phi_cut(self, i, levels, t_idx=0, ax=None, **kwargs):
        """
        Plot contours on a phi cut.

        Parameters
        ----------
        i : int
            Index at which to slice the data.
        levels : list
            List of levels to contour.
        t_idx : int, optional
            Time index at which to slice the data.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = viz.setup_polar_ax(ax)
        sliced = self.data.isel(phi=i, time=t_idx)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xr.plot.contour(sliced, x='theta', y='r', ax=ax,
                        levels=levels, **kwargs)
        viz.format_polar_ax(ax)
        ax.set_title(title)

    @property
    def _equator_theta_idx(self):
        """
        The theta index of the solar equator.
        """
        return (self.data.shape[1] - 1) // 2

    # Methods for equatorial cuts
    @add_common_docstring(returns_doc=returns_doc)
    def plot_equatorial_cut(self, t_idx=None, ax=None, **kwargs):
        """
        Plot an equatorial cut.

        Parameters
        ----------
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        t_idx : int, optional
            Time index at which to slice the data. If not given, an anmiation
            will be created across all time indices.
        kwargs :
            Additional keyword arguments are passed to
            `xarray.plot.pcolormesh`.

        Returns
        -------
        {returns_doc}
        """
        theta_slice = self.data.isel(theta=self._equator_theta_idx)
        time_slice = theta_slice.isel(time=t_idx or 0)

        ax = viz.setup_polar_ax(ax)
        kwargs = self._set_cbar_label(kwargs, self.unit.to_string('latex'))
        # Take slice of data and plot
        quad_mesh = time_slice.plot(x='phi', y='r', ax=ax, **kwargs)
        viz.format_equatorial_ax(ax)

        ax.set_title(f'{self.name}, equatorial plane')

        if t_idx is not None or self.n_timesteps == 1:
            return quad_mesh
        else:
            return viz.animate_time(ax, theta_slice, quad_mesh)

    def contour_equatorial_cut(self, levels, t_idx=0, ax=None, **kwargs):
        """
        Plot contours on an equatorial cut.

        Parameters
        ----------
        levels : list
            List of levels to contour.
        ax : matplolit.axes.Axes, optional
            axes on which to plot. Defaults to current axes if not specified.
        t_idx : int, optional
            Time index at which to slice the data.
        kwargs :
            Additional keyword arguments are passed to `xarray.plot.contour`.
        """
        ax = viz.setup_polar_ax(ax)
        sliced = self.data.isel(theta=self._equator_theta_idx, time=t_idx)
        # Need to save a copy of the title to reset it later, since xarray
        # tries to set it's own title that we don't want
        title = ax.get_title()
        xr.plot.contour(sliced, x='phi', y='r', ax=ax,
                        levels=levels, **kwargs)
        viz.format_equatorial_ax(ax)
        ax.set_title(title)

    @staticmethod
    def _set_cbar_label(kwargs, label):
        """
        Set the colobar label with units.
        """
        # Copy kwargs to prevent modifying them inplace
        kwargs = copy.deepcopy(kwargs)
        # Set the colobar label with units
        cbar_kwargs = kwargs.pop('cbar_kwargs', {})
        cbar_kwargs['label'] = cbar_kwargs.pop('label', label)
        kwargs['cbar_kwargs'] = cbar_kwargs
        return kwargs

    @u.quantity_input
    def sample_at_coords(self, lon: u.deg, lat: u.deg, r: u.m, t=None):
        """
        Sample this variable along a 1D trajectory of coordinates.

        Parameters
        ----------
        lon : astropy.units.Quantity
            Longitudes.
        lat : astropy.units.Quantity
            Latitudes.
        r : astropy.units.Quantity
            Radial distances.
        t : array-like, optional
            Timsteps. If the variable only has a single timstep, this argument
            is not required.

        Returns
        -------
        astropy.units.Quantity
            The sampled data.

        Notes
        -----
        Linear interpolation is used to interpoalte between cells. See the
        docstring of `scipy.interpolate.interpn` for more information.
        """
        dims = ['phi', 'theta', 'r', 'time']
        points = [self.data.coords[dim].values for dim in dims]
        values = self.data.values

        # Check that coordinates are increasing
        if not np.all(np.diff(points[1]) >= 0):
            raise RuntimeError(
                'Longitude coordinates are not monotonically increasing')
        if not np.all(np.diff(points[2]) >= 0):
            raise RuntimeError(
                'Latitude coordinates are not monotonically increasing')
        if not np.all(np.diff(points[3]) > 0):
            raise RuntimeError(
                'Radial coordinates are not monotonically increasing')

        # Pad phi points so it's possible to interpolate all the way from
        # 0 to 360 deg
        pcoords = points[0]
        pcoords = np.append(pcoords, pcoords[0] + 2 * np.pi)
        pcoords = np.insert(pcoords, 0, pcoords[-2] - 2 * np.pi)
        points[0] = pcoords

        values = np.append(values, values[0:1, :, :, :], axis=0)
        values = np.insert(values, 0, values[-2:-1, :, :, :], axis=0)

        if len(points[3]) == 1:
            # Only one timestep
            xi = np.column_stack([lon.to_value(u.rad),
                                  lat.to_value(u.rad),
                                  r.to_value(const.R_sun)])
            values = values[:, :, :, 0]
            points = points[:-1]
        else:
            xi = np.column_stack([lon.to_value(u.rad),
                                  lat.to_value(u.rad),
                                  r.to_value(const.R_sun),
                                  t])

        for i, dim in enumerate(dims[:-1]):
            bounds = np.min(points[i]), np.max(points[i])
            if not (np.all(bounds[0] <= xi[:, i]) and np.all(xi[:, i] <= bounds[1])):
                raise ValueError(f"At least one point is outside bounds {bounds} in {dim} dimension.")


        values_x = interpolate.interpn(points, values, xi)
        return values_x * self._unit



"""
Animating plots
===============
How to make animated plots.

If multiple timesteps are loaded, then it is possible to make an animation with
any of the plotting methods associated with a `Variable` object.
"""
###############################################################################
# First, load the required modules.
from psipy.data import sample_data
from psipy.model import MASOutput

###############################################################################
# Next, load a set of MAS output files. Here we just download two succesive
# Carrington rotations, but you can change this line to point to any directory
# that has data files spanning multiple timesteps of a simulation.
mas_path = sample_data.mas_helio_timesteps()
model = MASOutput(mas_path)

###############################################################################
# Get a variable from the model output. We can also inspect the variable to see
# how many timesteps it has.
vr = model['vr']
print(f'Number of timesteps: {vr.n_timesteps}')


###############################################################################
# We can now create an animation. In this example we plot a radial cut at the
# outer boundary of the model (i.e. a radial index of ``-1``).
animation = vr.plot_radial_cut(r_idx=-1)

###############################################################################
# Note that ``animation`` is a `matplotlib.animation.Animation` object. To save
# the animation to disk you can use the
# :meth:`~matplotlib.animation.Animation.save` method, e.g.:
# ``animation.save('mymovie.mp4')``




import astropy.units as u
import numpy as np

from psipy.data import sample_data
from psipy.model import MASOutput
from psipy.tracing import FieldLines, FortranTracer


def test_tracer():
    # Simple smoke test of field line tracing
    mas_path = sample_data.mas_sample_data()
    model = MASOutput(mas_path)
    bs = model.cell_corner_b()
    # Fake data to be unit vectors pointing in radial direction
    bs.loc[..., 'bp'] = 0
    bs.loc[..., 'bt'] = 0
    bs.loc[..., 'br'] = 1

    def cell_corner_b(self):
        return bs

    model.cell_corner_b = cell_corner_b

    tracer = FortranTracer()

    r = 40
    lat = 0 * u.deg
    lon = 0 * u.deg

    flines = tracer.trace(model, lon=lon, lat=lat, r=r)
    assert len(flines) == 1

    # Check that with auto step size, number of steps is close to number of
    # radial coordinates
    assert len(bs.coords['r']) == 140
    assert flines[0].xyz.shape == (139, 3)

    tracer = FortranTracer(step_size=0.5)
    flines = tracer.trace(model, lon=lon, lat=lat, r=r)
    # Check that changing step size has an effect
    assert flines[0].xyz.shape == (278, 3)


def test_fline_io(mas_model, tmpdir):
    # Simple smoke test of field line tracing
    mas_path = sample_data.mas_sample_data()
    model = MASOutput(mas_path)
    tracer = FortranTracer()

    r = 40
    lat = 0 * u.deg
    lon = 0 * u.deg

    flines = tracer.trace(model, lon=lon, lat=lat, r=r)
    fline_0 = flines[0]
    flines.save(tmpdir / 'flines.npz')
    del flines

    loaded_flines = FieldLines.load(tmpdir / 'flines.npz')
    fline_1 = loaded_flines[0]

    np.testing.assert_allclose(fline_0.r, fline_1.r)
    np.testing.assert_allclose(fline_0.lon, fline_1.lon)
    np.testing.assert_allclose(fline_0.lat, fline_1.lat)



import pytest
import xarray as xr

from psipy.io import mas
from psipy.model import MASOutput


def test_read_mas_error(tmp_path):
    with pytest.raises(FileNotFoundError):
        mas.read_mas_file(tmp_path, 'rho')

    with pytest.raises(FileNotFoundError):
        mas.get_mas_variables(tmp_path)


def test_read_mas_file(mas_directory):
    # Check that loading a single file works
    data = mas.read_mas_file(mas_directory, 'rho')
    assert isinstance(data, xr.Dataset)
    assert 'rho' in data


def test_save_netcdf(mas_directory):
    # Check that converting to netcdf works
    mas.convert_hdf_to_netcdf(mas_directory, 'rho')
    netcdf_dir = mas_directory / '..' / 'netcdf'

    netcdf_model = MASOutput(netcdf_dir)
    hdf_model = MASOutput(mas_directory)
    assert netcdf_model._data == hdf_model._data
